@@ -20,11 +20,11 @@
 #if CPPTOML_HAS_STD_REGEX
 #include <regex>
 #endif
+#include <map>
 #include <sstream>
 #include <stdexcept>
 #include <string>
 #include <unordered_map>
-#include <map>
 #include <vector>
 
 namespace cpptoml
@@ -66,6 +66,11 @@ class option
         return value_;
     }
 
+    const T* operator->() const
+    {
+        return &value_;
+    }
+
     const T& value_or(const T& alternative) const
     {
         if (!empty_)
@@ -166,27 +171,115 @@ struct valid_value
           || std::is_same<T, datetime>::value;
 };
 
-template <class T, bool Valid = valid_value<typename std::decay<T>::type>::value
-                                || std::is_convertible<T, std::string>::value>
+template <class T, class Enable = void>
 struct value_traits;
 
 template <class T>
-struct value_traits<T, true>
+struct valid_value_or_string_convertible
 {
-    const static bool valid = valid_value<typename std::decay<T>::type>::value
-                              || std::is_convertible<T, std::string>::value;
 
+    const static bool value = valid_value<typename std::decay<T>::type>::value
+                              || std::is_convertible<T, std::string>::value;
+};
+
+template <class T>
+struct value_traits<T, typename std::
+                           enable_if<valid_value_or_string_convertible<T>::
+                                         value>::type>
+{
     using value_type = typename std::
         conditional<valid_value<typename std::decay<T>::type>::value,
                     typename std::decay<T>::type, std::string>::type;
 
     using type = value<value_type>;
+
+    static value_type construct(T&& val)
+    {
+        return value_type{val};
+    }
+};
+
+template <class T>
+struct value_traits<T,
+                    typename std::
+                        enable_if<!valid_value_or_string_convertible<T>::value
+                                  && std::is_floating_point<
+                                         typename std::decay<T>::type>::value>::
+                            type>
+{
+    using value_type = typename std::decay<T>::type;
+
+    using type = value<double>;
+
+    static value_type construct(T&& val)
+    {
+        return value_type{val};
+    }
+};
+
+template <class T>
+struct value_traits<T,
+                    typename std::
+                        enable_if<!valid_value_or_string_convertible<T>::value
+                                  && std::is_signed<typename std::decay<T>::
+                                                        type>::value>::type>
+{
+    using value_type = int64_t;
+
+    using type = value<int64_t>;
+
+    static value_type construct(T&& val)
+    {
+        if (val < std::numeric_limits<int64_t>::min())
+            throw std::underflow_error{"constructed value cannot be "
+                                       "represented by a 64-bit signed "
+                                       "integer"};
+
+        if (val > std::numeric_limits<int64_t>::max())
+            throw std::overflow_error{"constructed value cannot be represented "
+                                      "by a 64-bit signed integer"};
+
+        return static_cast<int64_t>(val);
+    }
+};
+
+template <class T>
+struct value_traits<T,
+                    typename std::
+                        enable_if<!valid_value_or_string_convertible<T>::value
+                                  && std::is_unsigned<typename std::decay<T>::
+                                                          type>::value>::type>
+{
+    using value_type = int64_t;
+
+    using type = value<int64_t>;
+
+    static value_type construct(T&& val)
+    {
+        if (val > static_cast<uint64_t>(std::numeric_limits<int64_t>::max()))
+            throw std::overflow_error{"constructed value cannot be represented "
+                                      "by a 64-bit signed integer"};
+
+        return static_cast<int64_t>(val);
+    }
 };
 
 class array;
 class table;
 class table_array;
 
+template <class T>
+struct array_of_trait
+{
+    using return_type = option<std::vector<T>>;
+};
+
+template <>
+struct array_of_trait<array>
+{
+    using return_type = option<std::vector<std::shared_ptr<array>>>;
+};
+
 template <class T>
 inline std::shared_ptr<typename value_traits<T>::type> make_value(T&& val);
 inline std::shared_ptr<array> make_array();
@@ -346,15 +439,14 @@ std::shared_ptr<typename value_traits<T>::type> make_value(T&& val)
 {
     using value_type = typename value_traits<T>::type;
     using enabler = typename value_type::make_shared_enabler;
-    return std::make_shared<value_type>(enabler{}, std::forward<T>(val));
+    return std::make_shared<value_type>(
+        enabler{}, value_traits<T>::construct(std::forward<T>(val)));
 }
 
 template <class T>
 inline std::shared_ptr<value<T>> base::as()
 {
-    if (auto v = std::dynamic_pointer_cast<value<T>>(shared_from_this()))
-        return v;
-    return nullptr;
+    return std::dynamic_pointer_cast<value<T>>(shared_from_this());
 }
 
 // special case value<double> to allow getting an integer parameter as a
@@ -374,9 +466,7 @@ inline std::shared_ptr<value<double>> base::as()
 template <class T>
 inline std::shared_ptr<const value<T>> base::as() const
 {
-    if (auto v = std::dynamic_pointer_cast<const value<T>>(shared_from_this()))
-        return v;
-    return nullptr;
+    return std::dynamic_pointer_cast<const value<T>>(shared_from_this());
 }
 
 // special case value<double> to allow getting an integer parameter as a
@@ -480,14 +570,32 @@ class array : public base
         std::vector<std::shared_ptr<value<T>>> result(values_.size());
 
         std::transform(values_.begin(), values_.end(), result.begin(),
-                       [&](std::shared_ptr<base> v)
-                       {
-                           return v->as<T>();
-                       });
+                       [&](std::shared_ptr<base> v) { return v->as<T>(); });
 
         return result;
     }
 
+    /**
+     * Obtains a option<vector<T>>. The option will be empty if the array
+     * contains values that are not of type T.
+     */
+    template <class T>
+    inline typename array_of_trait<T>::return_type get_array_of() const
+    {
+        std::vector<T> result;
+        result.reserve(values_.size());
+
+        for (const auto& val : values_)
+        {
+            if (auto v = val->as<T>())
+                result.push_back(v->get());
+            else
+                return {};
+        }
+
+        return {std::move(result)};
+    }
+
     /**
      * Obtains an array of arrays. Note that elements may be nullptr
      * if they cannot be converted to a array.
@@ -497,8 +605,7 @@ class array : public base
         std::vector<std::shared_ptr<array>> result(values_.size());
 
         std::transform(values_.begin(), values_.end(), result.begin(),
-                       [&](std::shared_ptr<base> v) -> std::shared_ptr<array>
-                       {
+                       [&](std::shared_ptr<base> v) -> std::shared_ptr<array> {
                            if (v->is_array())
                                return std::static_pointer_cast<array>(v);
                            return std::shared_ptr<array>{};
@@ -609,8 +716,7 @@ class array : public base
     array() = default;
 
     template <class InputIterator>
-    array(InputIterator begin, InputIterator end)
-        : values_{begin, end}
+    array(InputIterator begin, InputIterator end) : values_{begin, end}
     {
         // nothing
     }
@@ -640,6 +746,28 @@ inline std::shared_ptr<array> make_element<array>()
     return make_array();
 }
 
+/**
+ * Obtains a option<vector<T>>. The option will be empty if the array
+ * contains values that are not of type T.
+ */
+template <>
+inline typename array_of_trait<array>::return_type
+array::get_array_of<array>() const
+{
+    std::vector<std::shared_ptr<array>> result;
+    result.reserve(values_.size());
+
+    for (const auto& val : values_)
+    {
+        if (auto v = val->as_array())
+            result.push_back(v);
+        else
+            return {};
+    }
+
+    return {std::move(result)};
+}
+
 class table;
 
 class table_array : public base
@@ -756,6 +884,72 @@ inline std::shared_ptr<table_array> make_element<table_array>()
     return make_table_array();
 }
 
+// The below are overloads for fetching specific value types out of a value
+// where special casting behavior (like bounds checking) is desired
+
+template <class T>
+typename std::enable_if<!std::is_floating_point<T>::value
+                            && std::is_signed<T>::value,
+                        option<T>>::type
+get_impl(const std::shared_ptr<base>& elem)
+{
+    if (auto v = elem->as<int64_t>())
+    {
+        if (v->get() < std::numeric_limits<T>::min())
+            throw std::underflow_error{
+                "T cannot represent the value requested in get"};
+
+        if (v->get() > std::numeric_limits<T>::max())
+            throw std::overflow_error{
+                "T cannot represent the value requested in get"};
+
+        return {static_cast<T>(v->get())};
+    }
+    else
+    {
+        return {};
+    }
+}
+
+template <class T>
+typename std::enable_if<!std::is_same<T, bool>::value
+                            && std::is_unsigned<T>::value,
+                        option<T>>::type
+get_impl(const std::shared_ptr<base>& elem)
+{
+    if (auto v = elem->as<int64_t>())
+    {
+        if (v->get() < 0)
+            throw std::underflow_error{"T cannot store negative value in get"};
+
+        if (static_cast<uint64_t>(v->get()) > std::numeric_limits<T>::max())
+            throw std::overflow_error{
+                "T cannot represent the value requested in get"};
+
+        return {static_cast<T>(v->get())};
+    }
+    else
+    {
+        return {};
+    }
+}
+
+template <class T>
+typename std::enable_if<!std::is_integral<T>::value
+                            || std::is_same<T, bool>::value,
+                        option<T>>::type
+get_impl(const std::shared_ptr<base>& elem)
+{
+    if (auto v = elem->as<T>())
+    {
+        return {v->get()};
+    }
+    else
+    {
+        return {};
+    }
+}
+
 /**
  * Represents a TOML keytable.
  */
@@ -918,10 +1112,7 @@ class table : public base
     {
         try
         {
-            if (auto v = get(key)->as<T>())
-                return {v->get()};
-            else
-                return {};
+            return get_impl<T>(get(key));
         }
         catch (const std::out_of_range&)
         {
@@ -939,10 +1130,7 @@ class table : public base
     {
         try
         {
-            if (auto v = get_qualified(key)->as<T>())
-                return {v->get()};
-            else
-                return {};
+            return get_impl<T>(get_qualified(key));
         }
         catch (const std::out_of_range&)
         {
@@ -950,6 +1138,69 @@ class table : public base
         }
     }
 
+    /**
+     * Helper function that attempts to get an array of values of a given
+     * type corresponding to the template parameter for a given key.
+     *
+     * If the key doesn't exist, doesn't exist as an array type, or one or
+     * more keys inside the array type are not of type T, an empty option
+     * is returned. Otherwise, an option containing a vector of the values
+     * is returned.
+     */
+    template <class T>
+    inline typename array_of_trait<T>::return_type
+    get_array_of(const std::string& key) const
+    {
+        if (auto v = get_array(key))
+        {
+            std::vector<T> result;
+            result.reserve(v->get().size());
+
+            for (const auto& b : v->get())
+            {
+                if (auto val = b->as<T>())
+                    result.push_back(val->get());
+                else
+                    return {};
+            }
+            return {std::move(result)};
+        }
+
+        return {};
+    }
+
+    /**
+     * Helper function that attempts to get an array of values of a given
+     * type corresponding to the template parameter for a given key. Will
+     * resolve "qualified keys".
+     *
+     * If the key doesn't exist, doesn't exist as an array type, or one or
+     * more keys inside the array type are not of type T, an empty option
+     * is returned. Otherwise, an option containing a vector of the values
+     * is returned.
+     */
+    template <class T>
+    inline typename array_of_trait<T>::return_type
+    get_qualified_array_of(const std::string& key) const
+    {
+        if (auto v = get_array_qualified(key))
+        {
+            std::vector<T> result;
+            result.reserve(v->get().size());
+
+            for (const auto& b : v->get())
+            {
+                if (auto val = b->as<T>())
+                    result.push_back(val->get());
+                else
+                    return {};
+            }
+            return {std::move(result)};
+        }
+
+        return {};
+    }
+
     /**
      * Adds an element to the keytable.
      */
@@ -1036,6 +1287,70 @@ class table : public base
     string_to_base_map map_;
 };
 
+/**
+ * Helper function that attempts to get an array of arrays for a given
+ * key.
+ *
+ * If the key doesn't exist, doesn't exist as an array type, or one or
+ * more keys inside the array type are not of type T, an empty option
+ * is returned. Otherwise, an option containing a vector of the values
+ * is returned.
+ */
+template <>
+inline typename array_of_trait<array>::return_type
+table::get_array_of<array>(const std::string& key) const
+{
+    if (auto v = get_array(key))
+    {
+        std::vector<std::shared_ptr<array>> result;
+        result.reserve(v->get().size());
+
+        for (const auto& b : v->get())
+        {
+            if (auto val = b->as_array())
+                result.push_back(val);
+            else
+                return {};
+        }
+
+        return {std::move(result)};
+    }
+
+    return {};
+}
+
+/**
+ * Helper function that attempts to get an array of arrays for a given
+ * key. Will resolve "qualified keys".
+ *
+ * If the key doesn't exist, doesn't exist as an array type, or one or
+ * more keys inside the array type are not of type T, an empty option
+ * is returned. Otherwise, an option containing a vector of the values
+ * is returned.
+ */
+template <>
+inline typename array_of_trait<array>::return_type
+table::get_qualified_array_of<array>(const std::string& key) const
+{
+    if (auto v = get_array(key))
+    {
+        std::vector<std::shared_ptr<array>> result;
+        result.reserve(v->get().size());
+
+        for (const auto& b : v->get())
+        {
+            if (auto val = b->as_array())
+                result.push_back(val);
+            else
+                return {};
+        }
+
+        return {std::move(result)};
+    }
+
+    return {};
+}
+
 std::shared_ptr<table> make_table()
 {
     struct make_shared_enabler : public table
@@ -1154,10 +1469,8 @@ class parser
         bool inserted = false;
         while (it != end && *it != ']')
         {
-            auto part = parse_key(it, end, [](char c)
-                                  {
-                                      return c == '.' || c == ']';
-                                  });
+            auto part = parse_key(it, end,
+                                  [](char c) { return c == '.' || c == ']'; });
 
             if (part.empty())
                 throw_parse_exception("Empty component of table name");
@@ -1195,11 +1508,11 @@ class parser
         // table already existed
         if (!inserted)
         {
-            auto is_value = [](const std::pair<const std::string&,
-                                               const std::shared_ptr<base>&>& p)
-            {
-                return p.second->is_value();
-            };
+            auto is_value
+                = [](const std::pair<const std::string&,
+                                     const std::shared_ptr<base>&>& p) {
+                      return p.second->is_value();
+                  };
 
             // if there are any values, we can't add values to this table
             // since it has already been defined. If there aren't any
@@ -1228,10 +1541,8 @@ class parser
         std::string full_ta_name;
         while (it != end && *it != ']')
         {
-            auto part = parse_key(it, end, [](char c)
-                                  {
-                                      return c == '.' || c == ']';
-                                  });
+            auto part = parse_key(it, end,
+                                  [](char c) { return c == '.' || c == ']'; });
 
             if (part.empty())
                 throw_parse_exception("Empty component of table array name");
@@ -1314,10 +1625,7 @@ class parser
     void parse_key_value(std::string::iterator& it, std::string::iterator& end,
                          table* curr_table)
     {
-        auto key = parse_key(it, end, [](char c)
-                             {
-                                 return c == '=';
-                             });
+        auto key = parse_key(it, end, [](char c) { return c == '='; });
         if (curr_table->contains(key))
             throw_parse_exception("Key " + key + " already present");
         if (*it != '=')
@@ -1364,10 +1672,7 @@ class parser
         }
 
         if (std::find_if(it, key_end,
-                         [](char c)
-                         {
-                             return c == ' ' || c == '\t';
-                         })
+                         [](char c) { return c == ' ' || c == '\t'; })
             != key_end)
         {
             throw_parse_exception("Bare key " + key
@@ -1375,10 +1680,7 @@ class parser
         }
 
         if (std::find_if(it, key_end,
-                         [](char c)
-                         {
-                             return c == '[' || c == ']';
-                         })
+                         [](char c) { return c == '[' || c == ']'; })
             != key_end)
         {
             throw_parse_exception("Bare key " + key
@@ -1510,64 +1812,60 @@ class parser
     {
         std::stringstream ss;
 
-        auto is_ws = [](char c)
-        {
-            return c == ' ' || c == '\t';
-        };
+        auto is_ws = [](char c) { return c == ' ' || c == '\t'; };
 
         bool consuming = false;
         std::shared_ptr<value<std::string>> ret;
 
         auto handle_line
-            = [&](std::string::iterator& it, std::string::iterator& end)
-        {
-            if (consuming)
-            {
-                it = std::find_if_not(it, end, is_ws);
+            = [&](std::string::iterator& it, std::string::iterator& end) {
+                  if (consuming)
+                  {
+                      it = std::find_if_not(it, end, is_ws);
 
-                // whole line is whitespace
-                if (it == end)
-                    return;
-            }
+                      // whole line is whitespace
+                      if (it == end)
+                          return;
+                  }
 
-            consuming = false;
+                  consuming = false;
 
-            while (it != end)
-            {
-                auto check = it;
-                // handle escaped characters
-                if (delim == '"' && *it == '\\')
-                {
-                    // check if this is an actual escape sequence or a
-                    // whitespace escaping backslash
-                    ++check;
-                    if (check == end)
-                    {
-                        consuming = true;
-                        break;
-                    }
+                  while (it != end)
+                  {
+                      auto check = it;
+                      // handle escaped characters
+                      if (delim == '"' && *it == '\\')
+                      {
+                          // check if this is an actual escape sequence or a
+                          // whitespace escaping backslash
+                          ++check;
+                          if (check == end)
+                          {
+                              consuming = true;
+                              break;
+                          }
 
-                    ss << parse_escape_code(it, end);
-                    continue;
-                }
+                          ss << parse_escape_code(it, end);
+                          continue;
+                      }
 
-                // if we can end the string
-                if (std::distance(it, end) >= 3)
-                {
-                    auto check = it;
-                    // check for """
-                    if (*check++ == delim && *check++ == delim
-                        && *check++ == delim)
-                    {
-                        it = check;
-                        ret = make_value<std::string>(ss.str());
-                        break;
-                    }
-                }
+                      // if we can end the string
+                      if (std::distance(it, end) >= 3)
+                      {
+                          auto check = it;
+                          // check for """
+                          if (*check++ == delim && *check++ == delim
+                              && *check++ == delim)
+                          {
+                              it = check;
+                              ret = make_value<std::string>(ss.str());
+                              break;
+                          }
+                      }
 
-                ss << *it++;
-            }
-        };
+                      ss << *it++;
+                  }
+              };
 
         // handle the remainder of the current line
         handle_line(it, end);
@@ -1669,16 +1967,14 @@ class parser
         // determine if we are an integer or a float
         auto check_it = it;
 
-        auto eat_sign = [&]()
-        {
+        auto eat_sign = [&]() {
             if (check_it != end && (*check_it == '-' || *check_it == '+'))
                 ++check_it;
         };
 
         eat_sign();
 
-        auto eat_numbers = [&]()
-        {
+        auto eat_numbers = [&]() {
             auto beg = check_it;
             while (check_it != end && is_number(*check_it))
             {
@@ -1774,11 +2070,9 @@ class parser
     std::shared_ptr<value<bool>> parse_bool(std::string::iterator& it,
                                             const std::string::iterator& end)
     {
-        auto boolend
-            = std::find_if(it, end, [](char c)
-                           {
-                               return c == ' ' || c == '\t' || c == '#';
-                           });
+        auto boolend = std::find_if(it, end, [](char c) {
+            return c == ' ' || c == '\t' || c == '#' || c == ',' || c == ']';
+        });
         std::string v{it, boolend};
         it = boolend;
         if (v == "true")
@@ -1792,12 +2086,10 @@ class parser
     std::string::iterator find_end_of_date(std::string::iterator it,
                                            std::string::iterator end)
     {
-        return std::find_if(it, end, [this](char c)
-                            {
-                                return !is_number(c) && c != 'T' && c != 'Z'
-                                       && c != ':' && c != '-' && c != '+'
-                                       && c != '.';
-                            });
+        return std::find_if(it, end, [this](char c) {
+            return !is_number(c) && c != 'T' && c != 'Z' && c != ':' && c != '-'
+                   && c != '+' && c != '.';
+        });
     }
 
     std::shared_ptr<value<datetime>>
@@ -1805,15 +2097,13 @@ class parser
     {
         auto date_end = find_end_of_date(it, end);
 
-        auto eat = [&](char c)
-        {
+        auto eat = [&](char c) {
             if (it == date_end || *it != c)
                 throw_parse_exception("Malformed date");
             ++it;
         };
 
-        auto eat_digits = [&](int len)
-        {
+        auto eat_digits = [&](int len) {
             int val = 0;
             for (int i = 0; i < len; ++i)
             {
@@ -1893,10 +2183,8 @@ class parser
             return make_array();
         }
 
-        auto val_end = std::find_if(it, end, [](char c)
-                                    {
-                                        return c == ',' || c == ']' || c == '#';
-                                    });
+        auto val_end = std::find_if(
+            it, end, [](char c) { return c == ',' || c == ']' || c == '#'; });
         parse_type type = determine_value_type(it, val_end);
         switch (type)
         {
@@ -1908,6 +2196,8 @@ class parser
                 return parse_value_array<double>(it, end);
             case parse_type::DATE:
                 return parse_value_array<datetime>(it, end);
+            case parse_type::BOOL:
+                return parse_value_array<bool>(it, end);
             case parse_type::ARRAY:
                 return parse_object_array<array>(&parser::parse_array, '[', it,
                                                  end);
